local lock = require("memory.rwlock")
local memory = import"memory"

ffi.cdef[[
    typedef struct ll ll_t;
    typedef struct ll_node ll_node_t;
    typedef void (*gen_fun_t)(void *);
    struct ll {
        int len;
        ll_node_t *hd;
        ll_node_t *nx;
        ll_node_t *tl;
        rw_lock_t m;
        gen_fun_t val_teardown;
    };
    typedef enum locktype locktype_t;
    enum locktype {
        l_read,
        l_write
    };
    struct ll_node {
        void *val;
        ll_node_t *nxt;
        rw_lock_t m;
    };
]]

local function ll_new(val_teardown)
    local list = ffi.cast("ll_t*", memory.alloc(ffi.sizeof("ll_t")))
    list.len = 0
    list.hd = nil
    list.nx = nil
    list.tl = nil
    lock.init(list.m)
    list.val_teardown = val_teardown
    return list
end

local function ll_delete(list)
    local node = list.hd
    local tmp
    while node ~= nil do
        if list.val_teardown ~= nil then
            list.val_teardown(node.val)
        end
        tmp = node
        node = node.nxt
        list.len = list.len - 1
    end
    list.hd = nil
    list.nx = nil
    list.tl = nil
    list.val_teardown = nil
    memory.free(list)
end

local function ll_new_node(val)
    local node = ffi.cast("ll_node_t*", memory.alloc(ffi.sizeof("ll_node_t")))
    node.val = ffi.cast("void*", val)
    node.nxt = nil
    lock.init(node.m)
    return node
end

local function ll_insert(list, val)
    local new_node = ll_new_node(val)
    lock.wlock(list.m)
    if list.nx ~= nil then
        lock.wlock(list.nx.m)
    end
    if list.tl ~= nil then
        if list.tl ~= list.nx then
            lock.wlock(list.tl.m)
        end
        list.tl.nxt = new_node
        if list.tl ~= list.nx then
            lock.unlock(list.tl.m)
        end
    else
        list.hd = new_node
    end
    list.tl = new_node
    list.len = list.len + 1
    if list.nx == nil then
        list.nx = new_node
    else
        lock.unlock(list.nx.m)
    end
    lock.unlock(list.m)
    return val
end

local function ll_lf_insert(list, val)
    local new_node = ll_new_node(val)
    if list.tl ~= nil then
        list.tl.nxt = new_node
    else
        list.hd = new_node
    end
    list.tl = new_node
    list.len = list.len + 1
    if list.nx == nil then
        list.nx = new_node
    end
    return val
end

local function ll_map(list, fun)
    lock.rlock(list.m)
    local node = list.hd
    lock.unlock(list.m)
    while node ~= nil do
        local m = node.m
        lock.wlock(m)
        local res = fun(node.val)
        if res then
            lock.unlock(m)
            return res
        end
        node = node.nxt
        lock.unlock(m)
    end
end

local function ll_lf_map(list, fun)
    local node = list.hd
    while node ~= nil do
        local m = node.m
        local res = fun(node.val)
        if res then
            return res
        end
        node = node.nxt
    end
end

local function ll_adv_nx(list)
    local nx
    lock.wlock(list.m)
    if list.nx ~= nil then
        nx = list.nx
        lock.wlock(nx.m)
        list.nx = list.nx.nxt
    elseif list.hd ~= nil then
        nx = list.hd
        lock.wlock(nx.m)
        list.nx = list.hd.nxt
    end
    lock.unlock(list.m)
    return nx
end

local function ll_map_adv(list, fun)
    local node = ll_adv_nx(list)
    while node ~= nil do
        local res = fun(node.val)
        if res then
            lock.unlock(node.m)
            return res
        end
        local tmp = node
        node = node.nxt
        if node ~= nil then
            lock.wlock(node.m)
        end
        lock.unlock(tmp.m)
    end
end

local function ll_len(list)
    lock.rlock(list.m)
    local len = list.len
    lock.unlock(list.m)
    return len
end

return {
    new = ll_new,
    delete = ll_delete,
    insert = ll_insert,
    lf_insert = ll_lf_insert,
    map = ll_map,
    lf_map = ll_lf_map,
    map_adv = ll_map_adv,
    len = ll_len
}