ffi.cdef[[
typedef struct __value{
    bool inf;
    unsigned long long index;
    void *value;
}bst_value_t;
typedef struct __tree{
    bst_value_t* value;
    struct __tree *left;
    struct __tree *right;
}bst_t;
typedef struct __seek_record{
    bst_t *ancestor;
    bst_t *successor;
    bst_t *parent;
    bst_t *leaf;
}bst_seek_record_t;
]]
local TAG_BIT = 1
local FLAG_BIT = 2

local bit = require("bit")

local function GET_ADDR(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return ffi.cast("bst_t*", bit.band(nodeptr, bit.bnot(bit.bor(TAG_BIT, FLAG_BIT))))
end
local function GET_LEFT(nodeptr)
    return GET_ADDR(nodeptr).left
end
local function GET_RIGHT(nodeptr)
    return GET_ADDR(nodeptr).right
end
local function GET_TAG(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return bit.band(nodeptr, TAG_BIT) ~= 0
end
local function GET_FLAG(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return bit.band(nodeptr, FLAG_BIT) ~= 0
end
local function TAGGED(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return ffi.cast("bst_t*", bit.bor(nodeptr, TAG_BIT), "bst_t*")
end
local function FLAGGED(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return ffi.cast("bst_t*", bit.bor(nodeptr, FLAG_BIT), "bst_t*")
end
local function UNTAGGED(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return ffi.cast("bst_t*", bit.band(nodeptr, bit.bnot(TAG_BIT)), "bst_t*")
end
local function UNFLAGGED(nodeptr)
    nodeptr = ffi.cast("int64_t", nodeptr)
    return ffi.cast("bst_t*", bit.band(nodeptr, bit.bnot(FLAG_BIT)), "bst_t*")
end

local function value_new(inf, idx, value)
    local v = ffi.cast("bst_value_t*", malloc(ffi.sizeof("bst_value_t")))
    v.inf = inf
    v.index = idx
    v.value = ffi.cast("void*", value)
    return v
end

local function node_new(inf, idx, value)
    local node = ffi.cast("bst_t*", malloc(ffi.sizeof("bst_t")))
    node.value = value_new(inf, idx, value)
    node.left = nil
    node.right = nil
    return node
end

local function tree_new()
    local R = node_new(true, 2, nil)
    local S = node_new(true, 1, nil)
    R.left = S
    R.right = node_new(true, 2, nil)
    S.left = node_new(true, 0, nil)
    S.right = node_new(true, 1, nil)
    return R
end

local function tree_free(tree)
    local nodes = {}
    table.insert(nodes, tree)
    local node
    while #nodes > 0 do
        node = table.remove(nodes)
        if node == nil then
            continue
        end
        if node.left then
            table.insert(nodes, GET_LEFT(node))
        end
        if node.right then
            table.insert(nodes, GET_RIGHT(node))
        end
        free(node)
    end
end

local function cleanup(value, record)
    local ancestor = record.ancestor
    local successor = record.successor
    local parent = record.parent
    local successor_addr
    local sibling_addr
    local child_addr
    if value < ancestor.value then
        successor_addr = "left"
    else
        successor_addr = "right"
    end
    if value < parent.value then
        child_addr = "left"
        sibling_addr = "right"
    else
        child_addr = "right"
        sibling_addr = "left"
    end
    if not GET_FLAG(parent[child_addr]) then
        sibling_addr = child_addr
    end
    print("CLEANUP PANIC")
    parent[sibling_addr] = TAGGED(parent[sibling_addr])
    local result
    if ancestor[successor_addr] == GET_ADDR(successor) then
        local tmp = ancestor[successor_addr]
        ancestor[successor_addr] = UNTAGGED(parent[sibling_addr])
        parent[sibling_addr] = tmp
        result = true
    else
        result = false
    end
    free(record)
    return result
end

local function tree_seek(tree, index)
    local record = ffi.cast("bst_seek_record_t*", malloc(ffi.sizeof("bst_seek_record_t")))
    record.ancestor = tree
    local S = tree
    record.successor = S
    record.parent = S
    record.leaf = GET_ADDR(GET_LEFT(S))
    local parentField = record.parent
    local currentField = record.leaf
    local current = GET_ADDR(currentField)
    while current ~= nil do
        if not GET_TAG(parentField) then
            record.ancestor = record.parent
            record.successor = record.leaf
        end
        record.parent = record.leaf
        record.leaf = current
        parentField = currentField
        if index < current.value.index then
            currentField = GET_LEFT(current)
        else
            currentField = GET_RIGHT(current)
        end
        current = GET_ADDR(currentField)
    end
    return record
end

local function tree_insert(tree, index, value)
    local record
    while true do
        record = tree_seek(tree, index)
        if record.leaf.value.index == index then
            return false
        else
            local parent = record.parent
            local leaf = record.leaf
            local new_leaf = node_new(false, index, value)
            local k = leaf.value
            local new_internal
            if k.index < index then
                new_internal = node_new(false, index, value)
                new_internal.left = leaf
                new_internal.right = new_leaf
            else
                new_internal = node_new(false, k.index, k.value)
                new_internal.left = new_leaf
                new_internal.right = leaf
            end
            local side
            if index < parent.value.index then
                side = "left"
            else
                side = "right"
            end
            if parent[side] == leaf then
                parent[side] = new_internal
                return true
            else
                free(new_leaf)
                free(new_internal)
                local addr = GET_ADDR(parent[side])
                if addr == leaf and (GET_TAG(parent[side]) or GET_FLAG(parent[side])) then
                    cleanup(v, record)
                end
            end
        end
    end
end

local function tree_remove(tree, index)
    local injecting = true
    local leaf = nil
    while true do
        local record = tree_seek(tree, index)
        local parent = record.parent
        local child_addr
        if index < parent.value.index then
            child_addr = "left"
        else
            child_addr = "right"
        end
        if injecting then
            leaf = record.leaf
            if leaf.value ~= v then
                return false
            end
            if parent[child_addr] == GET_ADDR(leaf) then
                parent[child_addr] = FLAGGED(UNTAGGED(leaf))
                injecting = false
                if cleanup(v, record) then
                    return true
                end
            end
        else
            if record.leaf ~= leaf or cleanup(v, record) then
                return true
            end
        end
    end
end

local function tree_contains(tree, index)
    local record = tree_seek(tree, index)
    local result = record.leaf.index == index
    free(record)
    return result
end

local function tree_inorder(tree)
    local out = {}
    local nodes = {}
    table.insert(nodes, tree)
    local current_node
    while #nodes > 0 do
        current_node = table.remove(nodes)
        if GET_ADDR(current_node.left) == nil then
            if not current_node.value.inf then
                table.insert(out, current_node.value)
            end
        else
            table.insert(nodes, current_node.right)
            table.insert(nodes, current_node.left)
        end
    end
    return out
end

local function tree_height(tree)
    local nodes = {}
    table.insert(nodes, tree)
    local current_node
    local max_height = 0
    local current_height = 0
    while #nodes > 0 do
        current_node = table.remove(nodes)
        if current_node.left == nil then
            if not current_node.value.inf then
                current_height = current_height + 1
                if current_height > max_height then
                    max_height = current_height
                end
            end
        else
            table.insert(nodes, current_node.right)
            table.insert(nodes, current_node.left)
            current_height = current_height + 1
        end
    end
    return max_height
end

return {
    new = tree_new,
    free = tree_free,
    insert = tree_insert,
    remove = tree_remove,
    contains = tree_contains,
    inorder = tree_inorder,
    height = tree_height
}