local rwlock = require("modules.memory.rwlock")
require("memory.new")

ffi.cdef[[
    typedef struct {
        size_t size;
        size_t capacity;
        rw_lock_t* lock;
        size_t element_size;
        char* __type;
        uint8_t** data; // uint8 so we can do pointer arithmetic
    } vector_t;
    typedef struct {
        float x, y;
    } vec2;
    typedef struct {
        float x, y, z;
    } vec3;
    typedef struct {
        float x, y, z, w;
    } vec4;
]]

register("vector", function(type)
    local v = ffi.cast("vector_t*", malloc(ffi.sizeof("vector_t")))
    v.size = 0
    v.lock = rwlock.new()
    local type_str = type .. "*"
    v.__type = malloc(#type_str + 1)
    ffi.copy(v.__type, type_str)
    v.element_size = ffi.sizeof(type)
    v.data = malloc(4 * v.element_size)
    v.capacity = 4
    return v
end)

local function cast(v)
    return ffi.cast("vector_t*", v)
end

local function delete(v)
    rwlock.delete(v.lock)
    free(v.data)
end

local function resize(v, size)
    rwlock.wlock(v.lock)
    if size > v.capacity then
        local n = realloc(v.data, size * v.element_size)
        if n ~= nil then
            v.data = n
        else
            rwlock.unlock(v.lock)
            error("Out of memory")
        end
        v.capacity = size
    end
    v.size = size
    rwlock.unlock(v.lock)
end

local function data__internal(v)
    return ffi.cast(ffi.string(v.__type), v.data)
end

local function push_back(v, value)
    rwlock.wlock(v.lock)
    if v.size >= v.capacity then
        local n = realloc(v.data, tonumber(v.capacity * 2 * v.element_size))
        if n ~= nil then
            v.data = n
        else
            rwlock.unlock(v.lock)
            error("Out of memory")
        end
        v.capacity = v.capacity * 2
    end
    local data = data__internal(v)
    if type(value) == "table" then
        local el = data[v.size]
        for k, e in pairs(value) do
            el[k] = e
        end
    else
        data[v.size] = value
    end
    v.size = v.size + 1
    rwlock.unlock(v.lock)
end

local function pop_back(v)
    rwlock.wlock(v.lock)
    if v.size > 0 then
        v.size = v.size - 1
    end
    rwlock.unlock(v.lock)
end

local function back(v)
    rwlock.rlock(v.lock)
    local ret = data__internal(v)[v.size - 1]
    rwlock.unlock(v.lock)
    return ret
end

local function front(v)
    rwlock.rlock(v.lock)
    local ret = data__internal(v)[0]
    rwlock.unlock(v.lock)
    return ret
end

local function at(v, index)
    rwlock.rlock(v.lock)
    local ret = data__internal(v)[index]
    rwlock.unlock(v.lock)
    return ret
end

local function clear(v)
    rwlock.wlock(v.lock)
    v.size = 0
    rwlock.unlock(v.lock)
end

local function empty(v)
    rwlock.rlock(v.lock)
    local ret = v.size == 0
    rwlock.unlock(v.lock)
    return ret
end

local function reserve(v, size)
    rwlock.wlock(v.lock)
    if size > v.capacity then
        local n = realloc(v.data, size * v.element_size)
        if n ~= nil then
            v.data = n
        else
            rwlock.unlock(v.lock)
            error("Out of memory")
        end
        v.capacity = size
    end
    rwlock.unlock(v.lock)
end

local function shrink_to_fit(v)
    rwlock.wlock(v.lock)
    if v.size < v.capacity then
        local n = realloc(v.data, v.size * v.element_size)
        if n ~= nil then
            v.data = n
        else
            rwlock.unlock(v.lock)
            error("Out of memory")
        end
        v.capacity = v.size
    end
    rwlock.unlock(v.lock)
end

local function insert(v, pos, value)
    rwlock.wlock(v.lock)
    if v.size >= v.capacity then
        local n = realloc(v.data, v.capacity * 2 * v.element_size)
        if n ~= nil then
            v.data = n
        else
            rwlock.unlock(v.lock)
            error("Out of memory")
        end
        v.capacity = v.capacity * 2
    end
    local data = data__internal(v)
    ffi.move(data + pos + 1, data + pos, (v.size - pos) * v.element_size)
    if type(value) == "table" then
        local el = data[pos]
        for k, e in pairs(value) do
            el[k] = e
        end
    else
        data[pos] = value
    end
    v.size = v.size + 1
    rwlock.unlock(v.lock)
end

local function erase(v, pos, count)
    count = count or 1
    rwlock.wlock(v.lock)
    if pos + count >= v.size then
        v.size = pos
    else
        local data = data__internal(v)
        ffi.move(data + pos, data + pos + count, (v.size - pos - count) * v.element_size)
        v.size = v.size - count
    end
    rwlock.unlock(v.lock)
end

local function erase_if(v, pred)
    rwlock.wlock(v.lock)
    local data = data__internal(v)
    local erase_start
    for i = tonumber(v.size) - 1, 0, -1 do
        local p = data[i]
        if pred(p) then
            if not erase_start then
                erase_start = i
            end
        elseif erase_start then
            local count = i - erase_start + 1
            ffi.move(data + erase_start, data + erase_start + count, (v.size - erase_start - count) * v.element_size)
            v.size = v.size - count
            erase_start = nil
        end
    end
    if erase_start then
        local count = erase_start + 1
        ffi.move(data, data + count, (v.size - count) * v.element_size)
        v.size = v.size - count
    end
    rwlock.unlock(v.lock)
end

local function swap(v1, v2)
    rwlock.wlock(v1.lock)
    rwlock.wlock(v2.lock)
    local t = v1.data
    v1.data = v2.data
    v2.data = t
    t = v1.size
    v1.size = v2.size
    v2.size = t
    t = v1.capacity
    v1.capacity = v2.capacity
    v2.capacity = t
    rwlock.unlock(v1.lock)
    rwlock.unlock(v2.lock)
end

local function size(v)
    return v.size
end

local function capacity(v)
    return v.capacity
end

local function data(v)
    rwlock.rlock(v.lock)
    local ret = ffi.cast(ffi.string(v.__type), v.data)
    rwlock.unlock(v.lock)
    return ret
end

return {
    cast = cast,
    delete = delete,
    size = size,
    capacity = capacity,
    resize = resize,
    push_back = push_back,
    pop_back = pop_back,
    back = back,
    front = front,
    at = at,
    clear = clear,
    empty = empty,
    reserve = reserve,
    shrink_to_fit = shrink_to_fit,
    data = data,
    insert = insert,
    erase = erase,
    erase_if = erase_if,
    swap = swap,
}